[comment {-*- tcl -*- doctools manpage}]
[manpage_begin adb n 1.0]
[moddesc {Tcl database}]
[titledesc {Database implementation for Tcl/Tk programs}]
[copyright {Roalt Aalmoes, 2007}]
[require Tcl 8]
[description]
[keywords database tcl mysql sql puretcl XML ]
[para]

This package provides an implementation of a database for Tcl/Tk
applications. The package is pure-Tcl, but it allows XML importing and
exporting facilities if the tdom package is available. The additional
package adbsql allows for an interface to mysql, via the tclmysql
package.

[para]

To use this package, make sure the source directory (where 
[file adb.tcl] and [file pkgIndex.tcl] are stored) is included in the 
[var auto_path] directory. Use can you the command lappend to add this
direcotry to your current path

[section {Context of the different commands}]

Here is a short summary of the main commands of ADB and how they are
used:

[list_begin bullet]

[bullet]
Always start your database by making calls to [cmd ADB::Init] to
initialize each table.
[bullet]
After that, you can fill the tables with calls to [cmd ADB::Add] (you
might find [cmd ADB::GetEmptyArray] useful here). Change records with
[cmd ADB::Change] and remove them with either [cmd ADB::Remove] or
[cmd ADB::RemoveKey].

[bullet]

To process individual records of the database, use calls to 
[cmd ADB::GetRecord] or [cmd ADB::GetArray]. If you have chosen to use
records, use the [cmd ADB::GetField] and [cmd ADB::SetField] to change
fields within the record.

To walk through all records, use the combination of 
[cmd ADB::FirstKey] and [cmd ADB::NextKey] commands.

[bullet]
To save your stored database, use calls 
to [cmd ADB::StoreCompositeFile] (ADB/Tcl own format) or 
[cmd ADB::StoreComposilteXMLFile] (for an XML-based format). To load
this file again, use [cmd ADB::ReadCompositeFile] or 
[cmd ADB::ReadCompositeXMLFile], respectively.
[bullet]

A collection of functions track or manipulate if changes are made to the
database. These functions are [cmd ADB::SetChanged], 
[cmd ADB::ChangedReset], [cmd ADB::Changed], 
[cmd ADB::AddGlobalListener], and [cmd ADB::RemoveGlobalListener].

[list_end]

[section COMMANDS]

[list_begin definitions]

[call [cmd ADB::AddGlobalListener]  [arg func ] ]

Add a global listener function [arg func]. This function will be
called if the database is changed and not yet written to storage. A
listener function may be useful to indicate to the user that changes
are not yet saved to non-volatile storage.

[call [cmd ADB::Add]  [arg table] [arg record ] ]

Add an array or record to the database. The [arg record] argument
should either be the content of a record or the content of an array
passed on as a list using the {array get arrayName}
expression. Internally is checked (based upon the size of the given
list) what way is used by the programmer.

[nl]
This function does not replace an already existing record.
[nl]
Returns 0 if everything ok or a string with a error message if something is wrong. 
[nl]See [cmd ADB::GetEmptyArray] for an example of [cmd ADB::Add] with an array.

[call [cmd ADB::ArrayToRecord]  [arg table] [arg arr ] ]

Converts an array to a record of the given [arg table] type. The array
must be passed on as a list using the [lb]array get arrayName[rb]
expression.

[call [cmd ADB::Change]  [arg table] [arg record ] ]


Change an array or record of the database. The [arg record] argument
should either be the content of a record or the content of an array
passed on as a list using the [lb]array get arrayName[rb]
expression. Internally is checked (based upon the size of the given
list) what way is used by the programmer.

[nl]
This function only replaces a record according to the primary fields.
[nl]
Returns 0 if everything ok or a string with a error message if something is wrong. 

[call [cmd ADB::ChangedReset] ]

Indicate that the internal database is saved. This function should be
called to indicate you have used one of the 
[cmd ADB::SaveCompositeFile] or [cmd ADB::SaveCompositeXMLFile] to save the
database. These two functions do not call this function internally!

[call [cmd ADB::Changed] ]

Returns boolean indicating of database has been changed, since it was
read or written to a file.  The value TRUE or 1 indicates that a
change occured.

[call [cmd ADB::Count]  [arg table ] ]  

Count the number of records in this table. If this table does not
exist or is empty, 0 is returned.

[nl]
[emph Example:]
[example_begin]
puts "Records in table Athletes: [lb]ADB::Count Athletes[rb]"
[example_end]

[call [cmd ADB::Empty] [arg table] ]

Empty the defined [arg table]. All data in the table is removed,
except for the defined structure.

[nl]

[emph Note] that the table should first be defined using the [cmd ADB::Init] command.

[call [cmd ADB::FirstKey]  [arg table ] ]  

Get for the first record the first unique, primary key for the given
[arg table]. After calling this function, use [cmd ADB::NextKey] to
get the subsequent next key.

[nl]
This function together with ADB::NextKey is used as an iterator.

[nl]

You can perform destructive action upon this record without changing
the rest of the iterators. The iteration records are determined during
this ADB::FirstKey call. However, if you add records, they will not be
included in the current iteration until you do a FirstKey call again.

[nl]
[emph Note] Expect a dictionary sorting order for the primary keys.
[nl]

[emph Note] You can use multiple FirstKey/NextKey operations at the
same time as long as they are for different tables.

[nl]
[emph Example:]
[example_begin]

proc ForAllAthletes { } {
    set atlkey [lb]ADB::FirstKey Athletes[rb]

    while { "$atlkey" != "" } {
        ADB::GetArray Athletes $atlkey athletearr
        
        # Do something with array athletearr, for instance:
        puts "Surname is $athletearr(surname)" 

        set atlkey [lb]ADB::NextKey Athletes[rb]
    }
}

[example_end]

[call [cmd ADB::GetArray]  [arg table] [arg recordKey] [arg arrayname ] ] 

Store the fields of the record indicated by [arg recordKey] of table
[arg table] into the array [arg arrayname]. Returns 0 if everything ok
or a string with a error message if something is wrong.

[nl]
[emph Note] See [cmd ADB::FirstKey] for an example.

[nl]

[emph Note] The array is passed on as a variable name.

[call [cmd ADB::GetEmptyArray]  [arg table] [arg arrayname ] ]

Store an empty record of the [arg table] into the array pointed to by
variable name [arg arrayname].


[nl]
[emph Example:]
[example_begin]
ADB::GetEmptyArray Athletes new_arr
set new_arr(no) 12445
set new_arr(firstname) John
set new_arr(surname) Doe
set new_arr(birthday) "01-01-2001"
ADB::Add Athletes [lb]array get new_arr[rb]
[example_end]

[call [cmd ADB::GetField]  [arg table] [arg field] [arg record ] ]

Get the value of a specified [arg field] within the given [arg record] of the named [arg table]. The record must be a list that contains the record, as it is gotten via a [cmd GetRecord] call.

[call [cmd ADB::GetRecord]  [arg table] [arg recordKey ] ]  

Return a record indicated by the [arg recordKey] of the given [arg table]. If the record cannot be found, an empty string is returned ("").

[call [cmd ADB::Headerline]  [arg table] [option {-ignoreheader list} ] ]  

This somewhat ackward function returned a text line containing the
descriptions of the fields as they are given via the [cmd ADB::Init]
command. The size of each field description is exactly as long as the
size of the given field length at the [cmd ADB::Init]. This function
may be useful to print as header line of a listing of the table
contents. Use the function [cmd ADB::Textline] to output a record as a
one line text line.

[nl]

The [arg {-ignoreheader list}] argument can be given to specify a list
of fields that will not be outputted.

[call [cmd ADB::Init] [arg table] [arg struc] [arg init_keyfunc]]

Initialise a [arg table] for the database by specifying the 
[arg struc] and providing a function that returns the primary key based
opon the given record.

[nl]

This command is [emph always] needed if you want to read or write a
database. The internal structure of the database is not stored in that
database, so even if your first action would be to read in a database,
you first have to specify this [cmd ADB::Init]. Note that this should
not be a big problem because in order to access the database, you (as
a programmer) must already have an understanding of the layout of the
database and the tables.

[nl]
The [emph structure] provided is a list containing a list for each field of the database.
The following fields are mandatory for each field:
[list_begin enum]
[enum]A unique field name that is used to refer to this field, for instance, if the record is stored as an array: the fields is then the index
[enum]A string describing the field. This string is used in [cmd ADB::Headerline].
[enum]A integer number indicating the size of the field. At this moment, this size is not always enforced when storing a field, but with MySQL connectivity it is and the length is truncated.
[enum]A value indicating if this is a primary field. Specify p for primary or just {} if not.
[enum]The type of the field. This can be [type alphanum] (alphanumeric), [type alpha] (alphabetical), [type int] (integer), [type date] (a date), [type time] (a time), or [type float] (a floating point). This field is mainly used to determine the right SQL datatype if you choose for SQL connectivity. How you fill in the actual field value should therefore be consistent in the way SQL stores these values.
[list_end]

[nl]
[emph Example:]
[example_begin]

proc InitAthletes { } {

    set structure { 
	{no        "Licence number" 10 p alphanum}
	{firstname "First name" 20 {} alphanum}
	{surname   "Surname" 20 {} alphanum}
        {sex       "Sex" 10 {} alpha}
        {birthday  "Birthday" 4 {} int}
	{club      "Club" 30 {} alphanum}
	{category  "Category" 10 {} alphanum}
    }
    ADB::Init Athletes $structure AthletesDBKey
}

proc AthletesDBKey { rec } {
    return [lb]lindex $rec 0[rb]
}

[example_end]

[call [cmd ADB::NextKey]  [arg table ] ]  

Get the next iteration key. Always first do a call to 
[cmd ADB::FirstKey] to initialise the iterator. See also this function for
further details and behaviour.

[call [cmd ADB::ReadCompositeFile]  [arg filename ] [opt "[arg "rootname"]"] [opt "[arg "parselogName"]"] ]  

Reads an ADB-specific file (written by the 
[cmd ADB::WriteCompositeFile] command) and stores it in the
database. 
[nl]

If the [arg rootname] is given, the root (or main element) of the XML
is compared to this value. Then, if they do not match, an error is
reported. If not given or if "" is specified, the main element name is
not checked and always accepted.

[nl]

If the [arg parselogName] is given, additional elements in the XML
file are reported here as a list of strings. For instance, if extra
elements, comments, etc. are encountered they are reported here. If
you want to enforce that no extra elements occur, specify the name of
a variable here and later check that the list stored in this variable
is empty.

[nl]

Returns either 0 if ok, or a string indicating an error.


[call [cmd ADB::ReadCompositeXMLFile]  [arg filename ] ]  


Reads the XML-formated ADB database file (written by the 
[cmd ADB::WriteCompositeXMLFile] command) and stores it in the
database. Returns either 0 if ok, or a string indicating an error.

[nl]

[emph Note] The tdom XML package should be installed for this function
to work, otherwise an error is returned.

[call [cmd ADB::RecordToArray]  [arg table] [arg arrayname] [arg record ] ]

Converts a database record to a database array given by the [arg arrayname] variable name.

[nl]
[emph Example:]
[example_begin]

set licence_no 12445
set rec [lb]ADB::GetRecord Athletes $licence_no[rb]
ADB::RecordToArray Athletes arr $rec
puts "Birthday is $arr(birthday)"

[example_end]

[call [cmd ADB::RemoveGlobalListener]  [arg func ] ]

Remove a previously added global listener call (via the 
[cmd ADB::AddGlobalListener]) from the list.

[call [cmd ADB::RemoveKey]  [arg table ] [arg recordkey ] ]

Remove record with given [arg recordkey]. Returns 0 is removed
successfully or a text string if an error occurred.

[call [cmd ADB::Remove]  [arg table] [arg record ] ]

Remove record specified by given record (or array passed on as a list
using the [lb]array get arrayName[rb] expression).

[call [cmd ADB::SetChanged] ]

Explicitly set that the database has been changed. If afterwards a
call is made to [cmd ADB::Changed], then TRUE is returned. Functions
within ADB that change the content of the database already call this
function internally (such as [cmd ADB::Change] or [cmd ADB::Add])

[call [cmd ADB::SetField]  [arg table] [arg field] [arg recordvar] [arg content ] ]

Replaces contents of field with [arg content] from [arg recordvar].

[nl]

[emph Note] This is one of the rare functions where the record is
passed on via the record variable name instead of the content of the
record. So specify the record name, not the content of the record.

[nl]
[emph Example:]
[example_begin]

set licence_no 12445
set rec [lb]ADB::GetRecord Athletes $licence_no[rb]
ADB::SetField Athletes club rec "Fortis"
ADB::Change Athletes $rec

[example_end]

[call [cmd ADB::StoreAsXML]  [arg filename] [arg rootname] [arg table ] ]  

Store table of database as XML file using the [arg rootname] as the
main (root) XML element. This function is also internally used to
store the complete database using [cmd DBA::StoreCompositeXMLFile].

[call [cmd ADB::StoreCompositeFile]  [arg filename] [arg listoftables ] ]  

Stores the database as a file that can be later be read via the 
[cmd ADB::ReadCompositeFile].

[call [cmd ADB::StoreCompositeXMLFile]  [arg filename] [arg rootname] [arg listoftables ] ]  

Writes the database to an XML-formatted file, that can later be read
by the [cmd ADB::ReadCompositeXMLFile]. Returns either 0 if ok, or a
string indicating an error.

[nl] [emph Note] The tdom XML package should be installed for this
function to work, otherwise an error is returned.

[call [cmd ADB::Structure]  [arg table] ]

Return the structure of the [arg table] in the database.

[call [cmd ADB::Textline]  [arg table] [arg recordkey] [option {-ignoreheader list} ] ]

This somewhat ackward function returned a text line containing the
content of a single record. The size of each field description is
exactly as long as the size of the given field length at the 
[cmd ADB::Init].  This function may be useful to print a listing of the
table contents. Use the function [cmd ADB::Headerline] to output a
header containing the description of each field.

[nl] The [arg {-ignoreheader list}] argument can be given to specify a
list of fields that will not be outputted.

[nl]
[emph Example:]
[example_begin]

proc table_to_listbox { scroll_frame table } {
    listbox $scroll_frame.list

    set i 0
    set objectptr [lb]ADB::FirstKey $table[rb]

    while { $objectptr != "" } {
	$scroll_frame.list insert end [lb]ADB::Textline $table $objectptr[rb]
	incr i
	set objectptr [lb]ADB::NextKey $table[rb] 
    }
    return $scroll_frame.list
}
[example_end]

[list_end]
[manpage_end]
