# ADB/ADBSQL - Tcl database Interface 
# Copyright (C) 2007 Roalt Aalmoes (adbsql@roalt.com)
#
# ADB - Tcl database Interface 
#
# This source code is dual-licensed under:
#  1. The LPGL (Lesser or Library GNU General Public license)
#  2. The New/Revisited BSD License.
#
# This means that you can pick either one of these license for your purpose of use.
# If you want to make improvements to this product and share them with the community,
# please provide them under both licenses to me.
# 

# For all procedures, except for functions: 0 if successful or a string containing
#                                           error message if failed.

package provide adb 1.0
package require adbsql
namespace eval ADB {

    # Following variables are arrays with index is tablename
    variable structure
    variable keyfunc
    variable no_fields
    variable primarykeys
    variable primarykeyidx
    variable NextList
    variable NextListElm

    # variable content_<name of table> is also array but has field name
    #                                  as index value

    # Global variables for all tables
    variable ChangedVar
    variable GlobalListeners
    variable XMLSupport 0

}

if [ catch { package require tdom } err ] {
    puts stdout "Cannot load tdom package, XML support turned off"
    set ADB::XMLSupport 0
} else {
    set ADB::XMLSupport 1
}

proc ADB::Init { table struc init_keyfunc } {
    variable structure 
    variable keyfunc 
    variable no_fields 
    variable primarykeys 
    variable primarykeyidx

    set structure($table) $struc
    set keyfunc($table) $init_keyfunc
    set no_fields($table) [llength $structure($table)]
    set primarykeys($table) {}
    set primarykeyidx($table) {}

    set i 0
    foreach istruc $structure($table) {
	set entryname [lindex $istruc 0]
        set primarykey [lindex $istruc 3]

        if { "$primarykey" == "p" } {
            lappend primarykeys($table) $entryname
            lappend primarykeyidx($table) $i
        }
        incr i
    }

    if { [ADBSQL::InUse] } {
        ADBSQL::CreateTable $table $struc $init_keyfunc
    } else {
        variable $table
        catch { unset $table }
    }
    return 0
}

proc ADB::Empty { table } {
    
    if { [ADBSQL::InUse] } {
        ADBSQL::SQLExec "DELETE FROM [ADBSQL::SQLTableName $table];"
    } else {
        variable content_$table
        catch { unset content_${table} }
    }
    ADB::SetChanged

    return 0
}

# purpose: Returns database structure as given in ADB::Init
# arguments: table - name of database
# returns: database structure of 'table'
proc ADB::Structure { table } {
    variable structure

    if { [info exist structure($table)] } {
	return $structure($table)
    } else {
	return {}
    }
}

# purpose: Indicate that database has been changed, so it
#          must be written to storage if changed need to
#          be saved.
proc ADB::SetChanged { } {
    variable ChangedVar
    variable GlobalListeners

    set ChangedVar 1

    if [info exist GlobalListeners] {
	foreach x $GlobalListeners {
	    $x
	}
    }
    
    return 0
}

# purpose: Register function that should be called if the database
#          is changed.
proc ADB::AddGlobalListener { func } {
    variable GlobalListeners
    
    if [info exist GlobalListeners] {
	lappend GlobalListeners $func
    } else {
	set GlobalListeners [list $func]
    }
 }

# purpose: Deregister function that should be called if the database
#          is changed.
proc ADB::RemoveGlobalListener { func } {
    variable GlobalListeners
    
    set newlistenerlist [list]
    foreach x $GlobalListeners {
	if { $x != $func } {
	    lappend newlistenerlist $x
	}
    }
    set GlobalListeners $newlistenerlist
}

# purpose: Indicate that database has just been written
#          to storage, so changes will not be lost
proc ADB::ChangedReset { } {
    variable ChangedVar

    set ChangedVar 0
    return 0
}

# purpose: Returns boolean that identifies whether
#          database has been changed since last
#          call of ADB::ChangedReset
proc ADB::Changed { } {
    variable ChangedVar

    return $ChangedVar
}


proc ADB::RemoveKey { table recordkey } {
    
    if { [ADBSQL::InUse] } {
        ADBSQL::RemoveKey $table $recordkey
        # For now, skip error reporting
    } else {
        variable content_${table}

        if { [info exists content_${table}($recordkey)] } {
            unset content_${table}($recordkey)
        } else {
            return "$recordkey does not exist"
        }
    }
    SetChanged
    
    return 0

}

proc ADB::Remove { table record } {
    variable keyfunc 
    variable no_fields
    
    if { [llength $record] == [expr 2 * $no_fields($table) ] } {
	# record is actually array, so convert
	set record [ArrayToRecord $table $record]
    } 

    set recordKey [$keyfunc($table) $record]

    return [RemoveKey $table $recordKey]
}

# purpose: add 'record' to table 'table'
proc ADB::Add { table record } {
    variable keyfunc 
    variable no_fields

    if { [llength $record] == [expr 2 * $no_fields($table) ] } {
	# record is actually array, so convert
	set record [ArrayToRecord $table $record]
    } elseif { [llength $record] != $no_fields($table) } {
        return "Received illegal record $record"
    }

    set recordKey [$keyfunc($table) $record]

    if { [info exists content_${table}($recordKey)] } {
	return "$recordKey already exists"
    } else {

        if { [ADBSQL::InUse] } {
            ADBSQL::Add $table $record
        } else {
            variable content_${table}

            set content_${table}($recordKey) $record
        }
        SetChanged

	return 0
    }
}

# purpose: change 'record' of table 'table'
proc ADB::Change { table record } {
    variable keyfunc 
    variable no_fields

    if { [llength $record] == [expr 2 * $no_fields($table) ] } {
	# record is actually array, so convert
	set record [ArrayToRecord $table $record]
    }

    set recordKey [$keyfunc($table) $record]

    if { [ADBSQL::InUse] } {
        ADBSQL::Change $table $record
    } else {
        variable content_$table
        
        if { [info exists content_${table}($recordKey)] } {
            set content_${table}($recordKey) $record
        } else {
            return "$recordKey does not exist"
        }
    }
    SetChanged
    
    return 0
}

# purpose : Zoek record met bepaalde key
# return: Record, or  "" if not found
proc ADB::GetRecord { table recordKey } {

    if { [ADBSQL::InUse] } {
        return [ADBSQL::GetRecord $table $recordKey]
    } else {
        variable content_$table

        if { [info exists content_${table}($recordKey)] } {
            return [set content_${table}($recordKey)]
        } else {
            return ""
        }
    }
}

proc ADB::GetArray { table recordKey arrayname} {
    variable structure

    upvar $arrayname return_array
    catch {unset return_array}

    set record ""

    if { [ADBSQL::InUse] } {
        set record [ADBSQL::GetRecord $table $recordKey]
    } else {
        variable content_$table 
        
        if { [info exists content_${table}($recordKey)] } {

            set record [GetRecord $table $recordKey]
	}
    }

    if { $record != "" } {
	# The same code as in ADB::RecordToArray

 	set i 0
	foreach istruc $structure($table) {
	    set return_array([lindex $istruc 0]) [lindex $record $i]
	    incr i
	}
	return 0
    } else {
	return "Record not found"
    }
}

proc ADB::GetEmptyArray { table arrayname} {
    variable structure

    upvar $arrayname return_array
    catch {unset return_array}
    
    set i 0
    foreach istruc $structure($table) {
        set return_array([lindex $istruc 0]) ""
        incr i
    }   
    return 0
}

proc ADB::RecordToArray { table arrayname record } {
    variable structure

    upvar $arrayname return_array

    catch {unset return_array}
    
    set i 0
    foreach istruc $structure($table) {
	set return_array([lindex $istruc 0]) [lindex $record $i]
	incr i
    }
    return 0
}

# return: Record
proc ADB::ArrayToRecord { table arr } {
    variable structure

    set rec {}

    foreach istruc $structure($table) {
	lappend rec [lindex $arr \
			 [expr [lsearch -exact $arr [lindex $istruc 0]] + 1] ]
    }
    return $rec
}

# purpose: Returns contents of field from record
# return: field or "" if not found
proc ADB::GetField { table field record } {
    variable structure

    set i 0
    foreach istruc $structure($table) {
	if { [lindex $istruc 0] == $field } {
	    return [lindex $record $i]
	} else {
	    incr i
	}
    }
    return ""
}

# purpose: Replaces contents of field with 'content' from record
# return: content or "" if not found
proc ADB::SetField { table field recordvar content } {
    variable structure
    
    upvar $recordvar record

    set i 0
    foreach istruc $structure($table) {
	if { [lindex $istruc 0] == $field } {
            set newrec [lreplace $record $i $i $content]
            set record $newrec
	    return $content
	} else {
	    incr i
	}
    }
    return ""
}

proc ADB::Textline { table recordkey args } {
    variable structure

    set arg(-ignoreheader) {}
    
    array set arg $args

    set record [GetRecord $table $recordkey]

    set i 0
    set textline ""
    set i_max [llength $structure($table)]
    
    for { set i 0 } { $i < $i_max } { incr i} {
	set istruc [lindex $structure($table) $i]
        set name [lindex $istruc 0]
        if { [lsearch $arg(-ignoreheader) $name] != -1 } {
            continue
        }
	set entrylengte [lindex $istruc 2]
	set formatstr [format "%%-%ds" $entrylengte]
	append textline " [format $formatstr [lindex $record $i]]|" 
    }
    return $textline
}

proc ADB::Headerline { table args } {
    variable structure

    set arg(-ignoreheader) {}
    
    array set arg $args

    set i 0
    set textline ""
    set i_max [llength $structure($table)]
    
    for { set i 0 } { $i < $i_max } { incr i} {
	set istruc [lindex $structure($table) $i]
        set name [lindex $istruc 0]
        if { [lsearch $arg(-ignoreheader) $name] != -1 } {
            continue
        }
	set entrylengte [lindex $istruc 2]
	set formatstr [format "%%-%d.%ds" $entrylengte $entrylengte]
	append textline " [format $formatstr $name]|" 
    }
    
    return $textline
}

proc ADB::Count { table } {
    if { [ADBSQL::InUse] } {
        return [ADBSQL::Count ($table)]
    } else {
        variable content_$table

        if { ! [array exists content_$table] } {
            return 0
        } else {
            return [array size content_$table]
        }
    }
}

# purpose: Returns sorted first key of record in (database 'table'
# returns: First record key or "" if table not found or no records
proc ADB::FirstKey { table } {
    variable NextList 
    variable NextListElm

    if { [ADBSQL::InUse] } {
        set NextList($table) [ADBSQL::PrimaryKeyList $table]
    } else {
        variable content_$table
     
        if { [array exists content_$table] != 1 } {
            return ""
        }
        set NextList($table) [lsort -dictionary [array names content_$table]]
    }
    set NextListElm($table) 0
    
    return [NextKey $table]
}

# purpose: Returns next sorted key of record in database 'table'.
#          ADB::FirstKey should be called prior to subsequent ADB::NextKey
#          calls.
# returns: next record key or "" if table not found or no record
# bugs: Bugs resolved
proc ADB::NextKey { table } {
    variable NextList 
    variable NextListElm

    if { $NextListElm($table) >= [llength $NextList($table) ] } {
	return ""
    } else {
	set nextkey [lindex $NextList($table) $NextListElm($table)]
	incr NextListElm($table)
	return $nextkey
    }
}

proc ADB::StoreFile { filename table } {

    if { [file exist "$filename"] && [file isfile "$filename"] } {
	catch {file delete "$filename"}
    }
    set fileId [open "$filename" w]
    set recordkey [FirstKey $table]

    while { $recordkey != "" } {
	set record [GetRecord $table $recordkey]
	puts $fileId $record
	set recordkey [NextKey $table]
    }

    close $fileId
    return 0
}

proc ADB::StoreAsXML { filename rootname table } {
    variable XMLSupport
    
    if { $XMLSupport == 0 } {
	return "XML Support is disabled (package tdom not found)"
    }
    if { [file exist "$filename"] && [file isfile "$filename"] } {
	catch {file delete "$filename"}
    }    
    set fileId [open "$filename" w]

    set doc [dom createDocument $rootname]
    StoreTable $doc $table

    puts $fileId [$doc asXML]
    $doc delete
    close $fileId
    return 0
}

# Translate
proc ADB::T { s } {
    return [TranslateTable::ToEnglish $s]
}

# Reverse
proc ADB::R { s } {
     return [TranslateTable::ToDutch $s]
}

proc ADB::StoreTable { doc  table } {

    set root [$doc documentElement]

    set recordkey [FirstKey $table]

    set table_structure [Structure $table]
    while { $recordkey != "" } {
	set record [GetRecord $table $recordkey]
	set tableElement [$doc createElement [T $table]]
	$root appendChild $tableElement
	set i 0
	foreach istruc $table_structure {
	    set tableField [$doc createElement [T [lindex $istruc 0]]]
	    $tableElement appendChild $tableField
	    $tableField appendChild [$doc createTextNode [lindex $record $i]]
	    incr i
	}
	set recordkey [NextKey $table]
    }
}

proc ADB::ReadFile { filename table } {
    set fileId [open "$filename" r]
    
    set errorcount 0
    while { [gets $fileId line] >= 0} {
	set result [Add $table $line]
	if { $result != 0 } {
	    incr errorcount
	}
    }
    if { $errorcount > 0 } {
	return "Could not read $errorcount lines"
    }
    return 0
}

proc ADB::ReadCompositeFile { filename } {

    set lineno 1
    
    if [catch {
	set fileId [open "$filename" r]
	
        set result2 0
	while { ( $result2 == 0 ) && [gets $fileId line] >= 0} {
	    set table [lindex $line 0]
	    set newline [lrange $line 1 end]
	    set result2 [Add $table $newline]
            
            incr lineno
	}
        close $fileId
    } result] {
	return "error reading line $lineno"
    } else {
        return $result2
    }
}

proc ADB::StoreCompositeFile { filename listoftables } {

    if { [file exist "$filename"] && [file isfile "$filename"] } {
	catch {file delete "$filename"}
    }
    if { [catch {set fileId [open "$filename" w]} result] } {
	return "Error opening writing $filename, error = $result"
    }
    if {[catch {

	foreach table $listoftables {
	    set recordkey [FirstKey $table]
	    
	    while { $recordkey != "" } {
		set record [GetRecord $table $recordkey]
		puts $fileId "$table $record"
		set recordkey [NextKey $table]
	    }
	}
	
    } result]} {
	close $fileId
	return "Error writing $result"
    } else {
	close $fileId

	return 0
    }
}

proc ADB::ReadCompositeXMLFile { filename { rootname "" } { parselogVar "" } } {
    variable XMLSupport
    
    if { "$parselogVar" != "" } {
	upvar $parselogVar parselog
    }

    set parselog {}
    
    if { $XMLSupport == 0 } {
	return "XML Support is disabled (package tdom not found)"
    }

    if [catch {
	set fileId [open "$filename" r]
    } result] {
	return "error reading file $filename"
    }
    if { [catch {dom parse [read $fileId]} doc] } {
	return "$doc"
    }
    set root     [$doc documentElement]

    if { "$rootname" != "" } {
	if { [$root nodeName] != "$rootname" } {
	    return "Wrong XML file: Main element [$root nodeName] does not match with expected $rootname"
	}
    }

    set table_node [$root firstChild]

    foreach table_node [$root childNodes] {
	if { "[$table_node nodeType]" == "ELEMENT_NODE" } {
	    set table [$table_node nodeName]
	    if { "$table" != "" } {
		set table [R $table]
	    }
	    if { [Structure $table] != {} } {
		# Table exists
		GetEmptyArray $table nodearr
		foreach field_node [$table_node childNodes] {
		    if { "[$field_node nodeType]" == "ELEMENT_NODE" } {
			set field [R [$field_node nodeName]]
			if { [info exist nodearr($field)] } {
			    set nodearr($field) [$field_node asText]
			} else {
			    lappend parselog "Not existing field:$field"
			}
		    } else {
			lappend parselog "Unused node type:[$field_node nodeType]"
		    }
		}
		set result [Add $table [array get nodearr]]
		if { $result != 0 } {
		    lappend parselog "Cannot add record:$result"
		    return "Cannot add record:$result"
		}
	    } else {
		lappend parselog "Element is not table:$table"
	    }
	} else {
	    lappend parselog "Unused node type:[$table_node nodeType]"
	}

    }
    return 0
}

proc ADB::StoreCompositeXMLFile { filename rootname listoftables } {
    variable XMLSupport
    
    if { $XMLSupport == 0 } {
	return "XML Support is disabled (package tdom not found)"
    }
    if { [file exist "$filename"] && [file isfile "$filename"] } {
	catch {file delete "$filename"}
    }
    if { [catch {set fileId [open "$filename" w]} result] } {
	return "Error opening writing $filename, error = $result"
    }
    if {[catch {
	set doc [dom createDocument $rootname]
	set root [$doc documentElement]

	foreach table $listoftables {
	    StoreTable $doc $table
	}
	puts $fileId {<?xml version="1.0" ?>}
	puts $fileId [$doc asXML]

	$doc delete

    } result]} {
	close $fileId
	return "Error writing $result"
    } else {
	close $fileId
	return 0
    }
}
