# ADB/ADBSQL - Tcl database Interface 
# Copyright (C) 2007 Roalt Aalmoes (adbsql@roalt.com)
#
# ADB - Tcl database Interface 
#
# This source code is dual-licensed under:
#  1. The LPGL (Lesser or Library GNU General Public license)
#  2. The New/Revisited BSD License.
#
# This means that you can pick either one of these license for your purpose of use.
# If you want to make improvements to this product and share them with the community,
# please provide them under both licenses to me.
# 

package provide adbsql 1.0

namespace eval ADBSQL {
    variable port 3306
    variable host ""
    variable user ""
    variable password ""
    variable database ""
    variable handler ""
    variable use_sql 0
    variable new_database 0
    variable SQLSupport 0
}

if [ catch { package require mysqltcl } err ] {
    puts stdout "Cannot load mysqltcl package, SQL support turned off"
    set ADBSQL::SQLSupport 0
} else {
    set ADBSQL::SQLSupport 1
}

proc ADBSQL::ErrorMessage { } {
    global mysqlstatus

    return $mysqlstatus(message)
}

# purpose: connect to MySQL database
# arguments: -port <portnumber>
#            -host <hostname of mysql database>
#            -user <mysql user name>
#            -password <mysql password of corresponding user>
#            -database <name of mysql database to use>
#            -new <true or false> : create a new database (true) or use existing one 
proc ADBSQL::Connect { args } {
    global mysqlstatus
    
    variable port
    variable host
    variable user
    variable password
    variable handler
    variable database
    variable use_sql
    variable new_database
    
    set arg(-port) 3306
    set arg(-host) 127.0.0.1
    set arg(-user) adbsql
    set arg(-password) adbsql
    set arg(-database) adbsql
    set arg(-new) false
    
    array set arg $args

    set host $arg(-host)
    set user $arg(-user)
    set password $arg(-password)
    set port $arg(-port)
    set database $arg(-database)
    
    if { $ADBSQL::SQLSupport == 0 } {
        puts stderr "Cannot open database: SQL support turned off, because mysqltcl package not available"
        return 0
    }

    if { [string tolower $arg(-new)] == "true" ||  $arg(-new) == "1" ||  [string tolower $arg(-new)] == "yes" } {
        set new_database 1
    }
    puts stderr "Connect to $host : $port for $user @ $password"

    if { [catch {mysqlconnect -host $host -port $port \
                     -user $user -password $password} handler] } {

        puts stderr "$mysqlstatus(message)"
        set handler ""
        set use_sql 0
        return 0
    }

    set use_sql 1
    
    if { $new_database } {
        set statement "CREATE DATABASE ${database};"
        if { ![ADBSQL::SQLExec $statement] } {
            ADBSQL::Disconnect
            set use_sql 0
            return 0
        }
    }
    
    if { [catch {mysqluse $handler $database} result] } {
        puts stderr "Cannot access database $database $mysqlstatus(message)"
        ADBSQL::Disconnect
        set use_sql 0
        return 0
    }

    return 1
}

proc ADBSQL::InUse { } {
    variable use_sql

    return $use_sql
}

proc ADBSQL::NewDatabase { } {
    variable new_database

    return $new_database
}

proc ADBSQL::Disconnect { } {
    variable handler
    variable use_sql

    if { "$handler" != "" } {
        catch {mysqlclose $handler}
        set handler ""
        set use_sql 0
    }
}

# purpose: Perform action using SQL, such as INSERT, DELETE, CREATE, DROP
# returns: 1 if ok or 0 in case of error
# note: In case of error use ADBSQL::ErrorMessage to get error
proc ADBSQL::SQLExec { statement } {
    variable handler
    global mysqlstatus
        
    if { [InUse] } {
        if { [catch {mysqlexec $handler $statement} result] } {
            puts stderr "SQLExec ERROR: $mysqlstatus(message) on $statement"
            return 0
        } else {
            return 1
        }
    }
    return 0
}

# purpose: Perform SQL SELECT statement
# returns: selected rows as a list or "0" if error occurred
# note: In case of error use ADBSQL::ErrorMessage to get error
proc ADBSQL::SQLSelect { statement } {
    variable handler
    global mysqlstatus
        
    if { [InUse] } {
        if { [catch {mysqlsel $handler $statement -list} result] } {
            puts stderr "$mysqlstatus(message)"
            return 0
        } else {
            return $result
        }
    }
    return 0
}

# purpose: Return table name as used in SQL database.
proc ADBSQL::SQLTableName { table } {
    return [string toupper [TranslateTable::ToEnglish $table]]
}

proc ADBSQL::CreateTable { table struc keyfunc } {
    global primarykeys$table primarykeyidx$table

    if { ![InUse] } {
        return 0
    }

    set primarykeys$table {}
    set primarykeyidx$table {}
    
    set englishtable [SQLTableName $table]

    if { [ADBSQL::NewDatabase] } {
        if { ! [SQLExec "DROP TABLE IF EXISTS $englishtable;"] } {
            return 0
        }
    }
    set statement "CREATE TABLE $englishtable ("

    set i 0
    foreach istruc [ADB::Structure $table] {
	set unconvertedentryname [lindex $istruc 0]
	set entrytext [lindex $istruc 1]
        set entrylength [lindex $istruc 2]
        set primarykey [lindex $istruc 3]
        set entryforce [lindex $istruc 4]

        set entryname [string toupper \
                           [TranslateTable::ToEnglish $unconvertedentryname]]
        
        if { $i != 0 } {
            append statement ", "
        }
        append statement "$entryname "
        
        switch -exact -- $entryforce {
            alphanum -
            alpha {
                append statement "VARCHAR($entrylength) "
            }
            int {
                 append statement "INTEGER "
            }
            date {
                append statement "VARCHAR(10) "
            }
            time {
                append statement "VARCHAR(10) "
            }
            float {
                append statement "DECIMAL(8,2) "
            }
            default {
                puts stderr "Unknown type $entryforce"
                append statement "VARCHAR(99) "
            }
        }
        if { "$primarykey" == "p" } {
            append statement "NOT NULL "
        }
        incr i
    }
    append statement ");"

    if { [ADBSQL::NewDatabase] } {
        SQLExec $statement
    }
}

# purpose: removes record with corresponding key
# returns: 1 if ok or 0 if failed
proc ADBSQL::RemoveKey { table recordkey } {
    global primarykeys$table

    if { ![InUse] } {
        return 0
    }

    set statement "DELETE FROM [SQLTableName $table] WHERE "
    
    set keys [split $recordkey &]

    if { [llength $keys] != [llength [set primarykeys$table]] } {
        puts stderr "Error, RemoveKey with invalid key for table $table"
        return
    }
    
    for { set i 0} { $i < [llength $keys] } { incr i } {
        set key [lindex $keys $i]
        set entryname [lindex [set primarykeys$table] $i]

        if { $i > 0 } {
            append statement "AND "
        }
        append statement "$entryname = '$key' "
        
    }
    append statement ";"

    SQLExec $statement
}

proc ADBSQL::Add { table record } {

    if { ![InUse] } {
        return 0
    }

    set statement "INSERT INTO [SQLTableName $table] VALUES ( "

    for { set i 0 } { $i < [llength $record] } { incr i } {
        
        if { $i != 0 } {
            append statement ", "
        }
        set value [lindex $record $i]
        append statement "'$value' "
    }
    append statement ");"

    SQLExec $statement
}

proc ADBSQL::Change { table record } {
    global primarykeys$table primarykeyidx$table

    if { ![InUse] } {
        return 0
    }
    
    set statement "UPDATE [SQLTableName $table] SET "

    for { set i 0 } { $i < [llength $record] } { incr i } {
        set istruc [lindex [ADB::Structure $table] $i]
        set unconvertedentryname [lindex $istruc 0]
        set entryname [string toupper \
                           [TranslateTable::ToEnglish $unconvertedentryname]]
        
        if { $i != 0 } {
            append statement ", "
        }
        set value [lindex $record $i]
        append statement "$entryname = '$value' "
    }
    append statement "WHERE "

    for { set i 0 } { $i < [llength [set primarykeyidx$table]] } { incr i } {
        if { $i != 0 } {
            append statement "AND "
        }
        set field [lindex [set primarykeys$table] $i]
        set idx [lindex [set primarykeyidx$table] $i]
        set value [lindex $record $idx]
        
        append statement "$field = '$value' "
    }

    append statement ";"

    SQLExec $statement
}

# returns: "" if not found
proc ADBSQL::GetRecord { table recordkey } {
    global primarykeys$table
    
    if { ![InUse] } {
        return ""
    }
    
    set statement "SELECT * FROM [SQLTableName $table] WHERE "

    set keys [split $recordkey &]

    if { [llength $keys] != [llength [set primarykeys$table]] } {
        puts stderr "Error, GetRecord with invalid key for table $table"
        return ""
    }
    
    for { set i 0} { $i < [llength $keys] } { incr i } {
        set key [lindex $keys $i]
        set entryname [lindex [set primarykeys$table] $i]

        if { $i > 0 } {
            append statement "AND "
        }
        append statement "$entryname = '$key' "
    }

    set result [SQLSelect $statement]
    if { "$result" != "0" && [llength $result] == 1 } {
        return [lindex $result 0]
    } else {
        return ""
    }
}

proc ADBSQL::Count { table } {
    set statement "SELECT COUNT(*) FROM [ADBSQL::SQLTableName $table];"
    set result [SQLSelect $statement]

    if { $result == 0 } {
        # error
        return 0
    } else {
        return [lindex $result 0]
    }
}

proc ADBSQL::PrimaryKeyList { table } {
    global primarykeys$table
    
    set statement "SELECT"
    
    set i 0
    foreach key [set primarykeys$table] {
        if { $i > 0 } {
            append statement ", "
        }
        append statement " $key"
        incr i
    }
    append statement " FROM [ADBSQL::SQLTableName $table];"
    
    set result [SQLSelect $statement]

    set resultlist {}
    if { $result != 0 } {
        foreach row $result {
            set key [join $row &]
            lappend resultlist $key
        }
    }

    return $resultlist
}
